/**
 * Title:		Graph
 * Copyright:	Copyright (c) Jan-Pascal van Best
 * Email:	    janpascal@vanbest.org
 * Author:	    Jan-Pascal van Best
 * Web: 		http://www.tbm.tudelft.nl/webstaf/janb/software/graph/api/index.html
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/**
 * Based on:
 * R. Davidson, D. Harel, "Drawing Graphs Nicely Using Simulated Annealing",
 * ACM Transactions on Graphics, Vol. 15, No. 4, October 1996, pp. 301-331
 */
package nl.tudelft.tbm.graph;


import java.io.*;
import java.lang.*;
import java.util.*;
import java.util.logging.*;

import java.awt.Frame;
import java.awt.geom.Line2D;


public class AnnealLayout implements Layout, Runnable {

    private Graph g;
    private GraphComponent c;
    private Thread runner = null;

    private double gamma = 0.85;	        // cooling factor
    private double startT = 0.001;		    // initial temperature
    private double endT = 0.000000005;        // final temperature
    private double nodeRepulsion = 1.0;		// node repulsion factor
    private double borderRepulsion = 1.0;   // border repulsion factor
    private double edgeStretch = 0.000001;       // edge stretch factor
    private double edgeCrossing = 1.0;      // edge crossing factor
    private double nodeEdgeRepulsion = 1.0; // node-edge repulsion factor
    private double edgeCrossCutoffT = 0.00000001; // where to start fine-tuning

    private double nodeEdgeMinimumSq;

    private double maxX = 100;
    private double maxY = 100;

    private double T;

    private Map inEdgesOfNode;
    private Map outEdgesOfNode;
    private Map edgesCrossingTestMap;

    private Random random;
    private static Logger logger = Logger.getLogger("nl.tudelft.tbm.graph");

    public AnnealLayout(Graph g, GraphComponent c) {
        this.g = g;
        this.c = c;
    }

    public AnnealLayout(Graph g) {
        this.g = g;
        this.c = null;
    }

    public AnnealLayout() {
        this.g = null;
        this.c = null;
    }

    public void setGraph(Graph g, GraphComponent c) {
        this.g = g;
        this.c = c;
    }

    public void setGamma(double gamma) {
        this.gamma = gamma;
    }

    public void setStartT(double startT) {
        this.startT = startT;
    }

    public void setEndT(double endT) {
        this.endT = endT;
    }

    public void setNodeRepulsion(double nodeRepulsion) {
        this.nodeRepulsion = nodeRepulsion;
    }

    public void setBorderRepulsion(double borderRepulsion) {
        this.borderRepulsion = borderRepulsion;
    }

    public void setEdgeStretch(double edgeStretch) {
        this.edgeStretch = edgeStretch;
    }

    public void setEdgeCrossing(double edgeCrossing) {
        this.edgeCrossing = edgeCrossing;
    }

    public void setNodeEdgeRepulsion(double nodeEdgeRepulsion) {
        this.nodeEdgeRepulsion = nodeEdgeRepulsion;
    }

    public void setEdgeCrossCutoffT(double edgeCrossCutoffT) {
        this.edgeCrossCutoffT = edgeCrossCutoffT;
    }

    public void doLayout() {
        if (this.runner == null) {
            this.runner = new Thread(this, "AnnealLayout");
            this.runner.start();
        }
    }

    public void interruptLayout() {
        if (this.runner != null) {
            runner.interrupt();
        }
    }

    public boolean isLayouting() {
        return (runner != null);
    }

    /* The following functions are only used in the runner thread */

    private double deltaENodeRepulsion(Node n, double newX, double newY) {
        if (nodeRepulsion == 0.0) return 0.0;

        double deltaE = 0.0;
        double distX;
        double distY;
        Node n2;

        for (Iterator i = g.nodes.iterator(); i.hasNext();) {
            n2 = (Node) i.next();
            if (n2 != n) {
                distX = (n2.getX() - newX);
                distY = (n2.getY() - newY);
                deltaE += 1 / ((distX * distX) + (distY * distY));

                distX = (n2.getX() - n.getX());
                distY = (n2.getY() - n.getY());
                deltaE -= 1 / ((distX * distX) + (distY * distY));
            }
        }

        return deltaE * nodeRepulsion;
    }

    private double deltaEBorderRepulsion(Node n, double newX, double newY) {
        if (borderRepulsion == 0.0) return 0.0;
        double x = n.getX();
        double y = n.getY();

        double oldE = 1 / (x * x) + 1 / (y * y) + 1 / (((maxX - x) * (maxX - x)) + ((maxY - y) * (maxY - y)));
        double newE = 1 / (newX * newX) + 1 / (newY * newY) + 1 / (((maxX - newX) * (maxX - newX)) + ((maxY - newY) * (maxY - newY)));

        return (newE - oldE) * borderRepulsion;
    }
    
    private double deltaEEdgeStretch(Node n, double newX, double newY) {
        if (edgeStretch == 0.0) return 0.0;
        List inEdges = (List) inEdgesOfNode.get(n);
        List outEdges = (List) outEdgesOfNode.get(n);

        double deltaX;
        double deltaY;

        double oldE;
        double newE;

        double x = n.getX();
        double y = n.getY();

        double deltaE = 0.0;

        for (Iterator i = inEdges.iterator(); i.hasNext();) {
            Edge e = (Edge) i.next();

            deltaX = ((Node) e.x).getX() - x;
            deltaY = ((Node) e.x).getY() - y;
            oldE = (deltaX * deltaX) + (deltaY * deltaY);
            deltaX = ((Node) e.x).getX() - newX;
            deltaY = ((Node) e.x).getY() - newY;
            newE = (deltaX * deltaX) + (deltaY * deltaY);
            deltaE += newE - oldE;
        }

        for (Iterator i = outEdges.iterator(); i.hasNext();) {
            Edge e = (Edge) i.next();

            deltaX = ((Node) e.y).getX() - x;
            deltaY = ((Node) e.y).getY() - y;
            oldE = (deltaX * deltaX) + (deltaY * deltaY);
            deltaX = ((Node) e.y).getX() - newX;
            deltaY = ((Node) e.y).getY() - newY;
            newE = (deltaX * deltaX) + (deltaY * deltaY);
            deltaE += newE - oldE;
        }

        return edgeStretch * deltaE;
    }
/*
    private static final boolean edgesCross(double x1, double y1, double x2, double y2, 
        double x3, double y3, double x4, double y4) {
//        return Line2D.linesIntersect(x1, y1, x2, y2, x3, y3, x4, y4);

        if (x1>x2) {
            double temp = x1; x1 = x2; x2 = temp;
            temp = y1; y1 = y2; y2 = temp;
        }

        if (x3>x4) {
            double temp = x3; x3 = x4; x4 = temp;
            temp = y3; y3 = y4; y4 = temp;
        }

        double alphaA = (y2-y1)/(x2-x1);
        double alphaB = (y4-y3)/(x4-x3);

        double betaA = y1-(alphaA*x1);
        double betaB = y3-(alphaB*x3);

        logger.info("    coords: "+x1+","+y1+"->"+x2+","+y2+" and "+x3+","+y3+"->"+x4+","+y4);
        logger.info("    alphaA: "+alphaA);
        logger.info("    alphaB: "+alphaB);
        logger.info("    x1==x2: "+(x1==x2));
        logger.info("    x3==x4: "+(x3==x4));
        logger.info("    beta A: "+betaA);
        logger.info("    beta B: "+betaB);

         // Check for vertical lines: alpha==+-Infinite

//        if (Double.isInfinite(alphaA)) { // x1==x2, first line vertical
//            if (Double.isInfinite(alphaB)) { // x3==x4
        if (x1==x2) { // alphaA is infinite, x1==x2, first line vertical
            if (x3==x4) { // alphaB is infinite, x3==x4
                // both lines vertical
                // FIXME: also check y's
                return (x1==x3); 
            } else {
                // only first line vertical
                if (! (x3<x1 && x1<x4) ) {
                    return false;
                }
                double y = alphaB*x1 + betaB;
                if (y3<y4) {
                    return (y3<y) && (y<y4);
                } else {
                    return (y4<y) && (y<y3);
                }
            }
//        } else if (Double.isInfinite(alphaB)) { // x3==x4
        } else if (x3==x4) { // alphaB is infinite, x3==x4
            // only second line vertical
            if (! (x1<x3 && x3<x2) ) {
                return false;
            }
            double y = alphaA*x3 + betaA;
            if (y1<y2) {
                return (y1<y) && (y<y2);
            } else {
                return (y2<y) && (y<y1);
            }
        }

        // No vertical lines from here

        double x = (betaB-betaA)/(alphaA-alphaB);
        boolean result = (x>x1) && (x<x2) && (x>x3) && (x<x4);
        
        if (true) {
            logger.info("         x: "+x);
            logger.info("edgesCross: "+result);
            try {
                BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
                br.readLine();
            } catch (IOException e) {
            }
        }
        
        return result;
      
    }
*/
    private static final boolean edgesCross(double x1, double y1, double x2, double y2, 
        double x3, double y3, double x4, double y4) {

        // P determines if the lines are parallel: lines are parallel iff. P==0
        double P = (x4-x3)*(y2-y1) -(x2-x1)*(y4-y3);

//        logger.info("    coords: "+x1+","+y1+"->"+x2+","+y2+" and "+x3+","+y3+"->"+x4+","+y4);
//        logger.info("         P: "+P);

        if (P==(double) 0.0) {
            return false; // Parallel lines don't cross, by definition ;)
        } else {
            double j = (x1*(y2-y3)+x2*(y3-y1)+x3*(y1-y2))/P;
            double i = (x1*(y4-y3)           +x3*(y1-y4)+x4*(y3-y1))/P;
/*
            logger.info("         i: "+i);
            logger.info("         j: "+j);
            try {
                BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
                br.readLine();
            } catch (IOException e) {
            }
*/
            return ( i>=0 && i<=1 && j>=0 && j<=1 );
        }

    }

    /**
     * Return the geometric distance between line segment AB and point P, with
     * (x1,y1) the coordinates of A; (x2,y2) the coordinates of B; and
     * (xp,yp) the coordinates of P
     */

    private static final double ptSegDistSq(double x1, double y1, double x2, double y2, double xp, double yp) {
        /* Segment through A and B is 
         * l: A + i(B-A)  i in [0,1]
         * direction perpendicular to  this segment is
         * V = (YA-YB, XA-XB)
         * Line though P, perpendicular to AB is
         * h: P + jV  j a real number
         * l crosses h: A+i(B-A) = P + jV
         * This leads to (for A!=B)
         *      (XP-XA)(XB-XA)+(YP-YA)(YB-YA)
         * i =  -----------------------------
         *          (XB-XA)^2 + (YB-YA)^2
         * And for the distance r:
         * if i in [0,1]: r^2 = |P-A-i(B-A)|^2
         * if i<0: r^2 = |P-A|^2
         * if i>1: r^2 = |P-B|^2
         *
         * For A=B, r^2 = |P-A|^2
         */
         if (x1==x2 && y1==y2) {
             return (yp-y1)*(yp-y1)+(xp-x1)*(xp-x1);
         }
         double i = ( (xp-x1)*(x2-x1) + (yp-y1)*(y2-y1) ) /
                    ( (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) );
         if (i<=0) {
             return (xp-x1)*(xp-x1) + (yp-y1)*(yp-y1);
         } else if (i>=1) {
             return (xp-x2)*(xp-x2) + (yp-y2)*(yp-y2);
         } else {
             double x = xp - x1 - i*(x2-x1);
             double y = yp - y1 - i*(y2-y1);
             return x*x + y*y;
         }
    }
    
    /* Distance between segments is either:
     *   - zero, of the segments intersect
     *   - or the minimum of the distances between 
     *     the endpoints of either segment  and
     *     the other edge
     */
    private static final double segSegDistSq(double x1, double y1, double x2, double y2, 
        double x3, double y3, double x4, double y4) {
        if (edgesCross(x1, y1, x2, y2, x3, y3, x4, y4)) {
            return 0.0;
        }
        double distanceSq = ptSegDistSq(x1, y1, x2, y2, x3, y3);

        double test = ptSegDistSq(x1, y1, x2, y2, x4, y4);
        if (test < distanceSq) distanceSq = test;

        test = ptSegDistSq(x3, y3, x4, y4, x1, y1);
        if (test < distanceSq) distanceSq = test;

        test = ptSegDistSq(x3, y3, x4, y4, x2, y2);
        if (test < distanceSq) distanceSq = test;

        return distanceSq;
    }

    private double deltaEEdgeCrossing(Node n, double newX, double newY) {
        if (edgeCrossing == 0.0) return 0.0;
        double x1;
        double y1;
        double x2;
        double y2;
        double x3;
        double y3;
        double x4;
        double y4;
        AbstractNode n1;
        AbstractNode n2;
        AbstractNode n3;
        AbstractNode n4;

        Object[] testEdges;

        List inEdges = (List) inEdgesOfNode.get(n);
        List outEdges = (List) outEdgesOfNode.get(n);

        int deltaE = 0;
        int oldCrossings = 0;
        int newCrossings = 0;

        //        logger.info("Testing inedges of "+n);
        for (Iterator i = inEdges.iterator(); i.hasNext();) {
            Edge e1 = (Edge) i.next();

            n1 = (AbstractNode) e1.x;
            n2 = (AbstractNode) e1.y; // n2==n !!
            x1 = n1.x; // n1.getX();
            y1 = n1.y; // n1.getY();
            x2 = n2.x; // n2.getX();
            y2 = n2.y; // n2.getY();
/*
            testEdges = new HashSet(g.edges);
            testEdges.removeAll((List) outEdgesOfNode.get(n1));
            testEdges.removeAll((List) inEdgesOfNode.get(n1));
            testEdges.removeAll(inEdges);
            testEdges.removeAll(outEdges);
*/
            testEdges = (Object[]) edgesCrossingTestMap.get(e1);

//            for (Iterator j = testEdges.iterator(); j.hasNext();) {
            for (int j = 0; j<testEdges.length; j++) {
                oldCrossings = 0;
                newCrossings = 0;
                Edge e2 = (Edge)testEdges[j];

                n3 = (AbstractNode) e2.x;
                n4 = (AbstractNode) e2.y;
                x3 = n3.x; // n3.getX();
                y3 = n3.y; // n3.getY();
                x4 = n4.x; // n4.getX();
                y4 = n4.y; // n4.getY();
                //                logger.info("Testing edges: "+e1+" and "+e2);
                //                logger.info("    "+x1+","+y1+"->"+newX+","+newY+" and "+x3+","+y3+"->"+x4+","+y4);
                if (edgesCross(x1, y1, newX, newY, x3, y3, x4, y4)) {
                    newCrossings++;
                    deltaE++;
                    //                    logger.info("    Moved edges cross: "+e1+" and "+e2);
                }
                //                logger.info("    edges: "+x1+","+y1+"->"+x2+","+y2+" and "+x3+","+y3+"->"+x4+","+y4);
                if (edgesCross(x1, y1, x2, y2, x3, y3, x4, y4)) {
                    oldCrossings++;
                    deltaE--;
                    //                    logger.info("    Edges cross: "+e1+" and "+e2);
                }

                /*
                 if (oldCrossings!=newCrossings) {
                 logger.info("old="+oldCrossings+"; new="+newCrossings);
                 logger.info("    Edges "+e1+" and "+e2);
                 logger.info("    Node "+n+" from "+n.getX()+","+n.getY()+" to "+newX+","+newY);
                 }
                 */
            }
        }

        //        logger.info("Testing outedges of "+n);
        for (Iterator i = outEdges.iterator(); i.hasNext();) {
            Edge e1 = (Edge) i.next();

            n1 = (AbstractNode) e1.x; // n1 == n !!
            n2 = (AbstractNode) e1.y;
            x1 = n1.x; // n1.getX();
            y1 = n1.y; // n1.getY();
            x2 = n2.x; // n2.getX();
            y2 = n2.y; // n2.getY();
/*
            testEdges = new HashSet(g.edges);
            testEdges.removeAll((List) outEdgesOfNode.get(n2));
            testEdges.removeAll((List) inEdgesOfNode.get(n2));
            testEdges.removeAll(inEdges);
            testEdges.removeAll(outEdges);
*/
//            testEdges = (Set) edgesCrossingTestSetMap.get(e1);
//            for (Iterator j = testEdges.iterator(); j.hasNext();) {
            testEdges = (Object[]) edgesCrossingTestMap.get(e1);

            for (int j = 0; j<testEdges.length; j++) {
                oldCrossings = 0;
                newCrossings = 0;
//                Edge e2 = (Edge) j.next();
                Edge e2 = (Edge)testEdges[j];

                n3 = (AbstractNode) e2.x;
                n4 = (AbstractNode) e2.y;
                x3 = n3.x; // n3.getX();
                y3 = n3.y; // n3.getY();
                x4 = n4.x; // n4.getX();
                y4 = n4.y; // n4.getY();
                //                logger.info("Testing edges: "+e1+" and "+e2);
                //                logger.info("    "+newX+","+newY+"->"+x2+","+y2+" and "+x3+","+y3+"->"+x4+","+y4);
                if (edgesCross(newX, newY, x2, y2, x3, y3, x4, y4)) {
                    newCrossings++;
                    deltaE++;
                    //                    logger.info("    moved edges cross: "+e1+" and "+e2);
                }
                //                logger.info("    "+x1+","+y1+"->"+x2+","+y2+" and "+x3+","+y3+"->"+x4+","+y4);
                if (edgesCross(x1, y1, x2, y2, x3, y3, x4, y4)) {
                    oldCrossings++;
                    deltaE--;
                    //                    logger.info("    edges cross: "+e1+" and "+e2);
                }

                /*
                 if (oldCrossings!=newCrossings) {
                 logger.info("old="+oldCrossings+"; new="+newCrossings);
                 logger.info("    Edges "+e1+" and "+e2);
                 logger.info("    Node "+n+" from "+n.getX()+","+n.getY()+" to "+newX+","+newY);
                 }
                 */
            }
        }

        if (deltaE != 0) {
            //            logger.info("Edge crossing: "+n.toString()+": "+deltaE);
        }

        return edgeCrossing * deltaE;
    }

    private final double nodeEdge(double distanceSq) {
        if (distanceSq < nodeEdgeMinimumSq) {
            return edgeCrossing;
        } else {
            return nodeEdgeRepulsion / distanceSq;
        }
    }


    private double deltaENodeEdgeRepulsion(Node nNode, double newX, double newY) {
        AbstractNode n = (AbstractNode) nNode;
        if (nodeEdgeRepulsion == 0.0) return 0.0;

//        logger.info("Testing nodeEdgeRepulsion for node "+n+" to "+newX+","+newY);

        double deltaE = 0;

        List inEdges = (List) inEdgesOfNode.get(n);
        List outEdges = (List) outEdgesOfNode.get(n);

        Set testEdges = new HashSet(g.edges);

        testEdges.removeAll(inEdges);
        testEdges.removeAll(outEdges);

        // First situation: distance from this node to other edges changes
        for (Iterator i = testEdges.iterator(); i.hasNext();) {
            Edge e = (Edge) i.next();
            AbstractNode n1 = (AbstractNode) e.x;
            AbstractNode n2 = (AbstractNode) e.y;
            double oldDistanceSq = ptSegDistSq(n1.x, n1.y, n2.x, n2.y, n.x, n.y);
            double newDistanceSq = ptSegDistSq(n1.x, n1.y, n2.x, n2.y, newX, newY);
//            logger.info("Comparing node "+n+" to edge "+e+": distance = "+Math.sqrt(oldDistanceSq)+"; new "+Math.sqrt(newDistanceSq));

            deltaE += nodeEdge(newDistanceSq) - nodeEdge(oldDistanceSq);
            
        }

        // Second situation: distance edge of which this node is the origin to another node or edge
//        logger.info("outEdges: "+outEdges);
        for (Iterator i = outEdges.iterator(); i.hasNext();) {
            Edge e = (Edge) i.next();
            // n1 == n1
            AbstractNode n2 = (AbstractNode) e.y; // dest node of edge

            // 2a: distance to other node
//            logger.info("nodes: "+g.nodes);
            for (Iterator j = g.nodes.iterator(); j.hasNext();) {
                AbstractNode otherNode = (AbstractNode) j.next();

                if (n == otherNode || n2 == otherNode) continue; // skip self and other endpoint of edge
                double oldDistanceSq = ptSegDistSq(n.x, n.y, n2.x, n2.y, otherNode.x, otherNode.y);
                double newDistanceSq = ptSegDistSq(newX, newY, n2.x, n2.y, otherNode.x, otherNode.y);
//                logger.info("Comparing "+e+" to node "+otherNode+": distance = "+Math.sqrt(oldDistanceSq)+"; new "+Math.sqrt(newDistanceSq));

                deltaE += nodeEdge(newDistanceSq) - nodeEdge(oldDistanceSq);
            }
            // 2b: distance to other edge
//            logger.info("testEdges: "+testEdges);
            for (Iterator j = testEdges.iterator(); j.hasNext();) {
                Edge otherEdge = (Edge) j.next();

                if (e.y == otherEdge.x || e.y == otherEdge.y) continue; // skip edges attached to this edge
                AbstractNode n3 = (AbstractNode) otherEdge.x; 
                AbstractNode n4 = (AbstractNode) otherEdge.y;
//                logger.info("Comparing "+e+" to "+otherEdge);
                double oldDistanceSq = segSegDistSq(n.x, n.y, n2.x, n2.y, n3.x, n3.y, n4.x, n4.y);
                double newDistanceSq = segSegDistSq(newX, newY, n2.x, n2.y, n3.x, n3.y, n4.x, n4.y);
//                logger.info("Comparing "+e+" to "+otherEdge+": distance = "+Math.sqrt(oldDistanceSq)+"; new "+Math.sqrt(newDistanceSq));

                deltaE += nodeEdge(newDistanceSq) - nodeEdge(oldDistanceSq);
            }
        }

//        logger.info("inEdges: "+inEdges);
        // Third situation: distance edge of which this node is the destination to another node or edge
        for (Iterator i = inEdges.iterator(); i.hasNext();) {
            Edge e = (Edge) i.next();
            AbstractNode n1 = (AbstractNode) e.x; // dest node of edge

            // 3a: distance to other node
//            logger.info("nodes: "+g.nodes);
            for (Iterator j = g.nodes.iterator(); j.hasNext();) {
                AbstractNode otherNode = (AbstractNode) j.next();

                if (n == otherNode || n1 == otherNode) continue; // skip self and other endpoint of edge
                double oldDistanceSq = ptSegDistSq(n1.x, n1.y, n.x, n.y, otherNode.x, otherNode.y);
                double newDistanceSq = ptSegDistSq(n1.x, n1.y, newX, newY, otherNode.x, otherNode.y);
//                logger.info("Comparing "+e+" to node "+otherNode+": distance = "+Math.sqrt(oldDistanceSq)+"; new "+Math.sqrt(newDistanceSq));

                deltaE += nodeEdge(newDistanceSq) - nodeEdge(oldDistanceSq);
            }
            // 3b: distance to other edge
//            logger.info("testEdges: "+testEdges);
            for (Iterator j = testEdges.iterator(); j.hasNext();) {
                Edge otherEdge = (Edge) j.next();

                if (e.x == otherEdge.x || e.x == otherEdge.y) continue; // skip edges attached to this edge
                AbstractNode n3 = (AbstractNode) otherEdge.x;
                AbstractNode n4 = (AbstractNode) otherEdge.y;
//                logger.info("Comparing "+e+" to "+otherEdge);
                double oldDistanceSq = segSegDistSq(n1.x, n1.y, n.x, n.y, n3.x, n3.y, n4.x, n4.y);
                double newDistanceSq = segSegDistSq(n1.x, n1.y, newX, newY, n3.x, n3.y, n4.x, n4.y);
//                logger.info("Comparing "+e+" to "+otherEdge+": distance = "+Math.sqrt(oldDistanceSq)+"; new "+Math.sqrt(newDistanceSq));

                deltaE += nodeEdge(newDistanceSq) - nodeEdge(oldDistanceSq);
            }
        }

        return deltaE;
    }
    
    private double deltaE(Node n, double newX, double newY) {
        return (deltaENodeRepulsion(n, newX, newY) +
                deltaEBorderRepulsion(n, newX, newY) +
                deltaEEdgeStretch(n, newX, newY) +
                ((T > edgeCrossCutoffT) ?
                    deltaEEdgeCrossing(n, newX, newY) :
                    deltaENodeEdgeRepulsion(n, newX, newY)));
    }

    private void tryAndMove(Node n) {
        //        double deltaX = ((random.nextDouble()*1.0)-0.5)*Math.sqrt(nodeRepulsion/T);
        //        double deltaY = ((random.nextDouble()*1.0)-0.5)*Math.sqrt(nodeRepulsion/T);
        //        double deltaX = (random.nextDouble()-0.5)*maxX;
        //        double deltaY = (random.nextDouble()-0.5)*maxY;

        double deltaX = 2 * (random.nextDouble() - 0.5) * Math.sqrt(0.5 * T / edgeStretch);
        double deltaY = 2 * (random.nextDouble() - 0.5) * Math.sqrt(0.5 * T / edgeStretch);

        if (Math.abs(deltaX) > maxX) deltaX = (random.nextDouble() - 0.5) * maxX;
        if (Math.abs(deltaY) > maxY) deltaY = (random.nextDouble() - 0.5) * maxY;

        double newX = n.getX() + deltaX;
        double newY = n.getY() + deltaY;

        // Do it this way because |deltaX|<=0.5 maxX
        // Now we "bounce back" if we hit the wall
        if (newX < 0) newX = Math.abs(newX); // maxX/1000*(1+random.nextDouble());
        if (newY < 0) newY = Math.abs(newY); // maxY/1000*(1+random.nextDouble());
        if (newX > maxX) newX = maxX - (newX - maxX); // maxX-Math.abs(maxX-(maxX/1000*(1+random.nextDouble()));
        if (newY > maxY) newY = maxY - (newY - maxY); // maxY-(maxY/1000*(1+random.nextDouble()));

        double deltaE = deltaE(n, newX, newY);

//        logger.info("Trying to move "+n+" to "+newX+","+newY+", deltaE="+deltaE);

        if (deltaE < 0 || random.nextDouble() < Math.exp(-deltaE / T)) {
            n.setPosition(newX, newY);
            // Tijdelijk!!!
//            c.doLayout();
//            c.repaint();

            //            if (Math.abs(deltaE)>5000) {

/*
            logger.info("Moved node "+n+" to "+newX+","+newY+" for deltaE="+deltaE);
            for(Iterator i=g.nodes.iterator(); i.hasNext(); ){
                Node node = (Node)i.next();
                logger.info("Node "+node+": "+node.getX()+","+node.getY());
            }
            try {
                BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
                br.readLine();
            } catch (IOException e) {}
*/            
        }
    }

    public void prepareEdgesOfNodeMap() {
        inEdgesOfNode = new HashMap();
        outEdgesOfNode = new HashMap();
        for (Iterator i = g.nodes.iterator(); i.hasNext();) {
            Node n = (Node) i.next();

            inEdgesOfNode.put(n, new ArrayList());
            outEdgesOfNode.put(n, new ArrayList());
        }

        for (Iterator i = g.edges.iterator(); i.hasNext();) {
            Edge e = (Edge) i.next();

            ((List) inEdgesOfNode.get(e.y)).add(e);
            ((List) outEdgesOfNode.get(e.x)).add(e);
        }
    }

    public void prepareEdgesCrossingMap() {
        edgesCrossingTestMap = new HashMap();
        for (Iterator i = g.edges.iterator(); i.hasNext(); ) {
            Edge e = (Edge) i.next();
            Node n1 = (Node) e.x;
            Node n2 = (Node) e.y;

            Set testEdges = new HashSet(g.edges);
            testEdges.removeAll((List) inEdgesOfNode.get(n1));
            testEdges.removeAll((List) outEdgesOfNode.get(n1));
            testEdges.removeAll((List) inEdgesOfNode.get(n2));
            testEdges.removeAll((List) outEdgesOfNode.get(n2));
            edgesCrossingTestMap.put(e, testEdges.toArray());
        }
    }


    public void run() {
        random = new Random(42);

        T = startT;

        // Calculate maxX, maxY
        if (c != null) {
            maxX = (c.getWidth() - (2 * c.NODE_SIZE) - 50) / (c.factorX);
            maxY = (c.getHeight() - (2 * c.NODE_SIZE) - 50) / (c.factorY);
        } else {
            maxX = 0;
            maxY = 0;
            for (Iterator i = g.nodes.iterator(); i.hasNext();) {
                Node n = (Node) i.next();

                if (n.getX() > maxX) maxX = n.getX();
                if (n.getY() > maxY) maxY = n.getY();
            }
            maxX *= 1.1;
            maxY *= 1.1;
        }

        nodeEdgeMinimumSq = nodeEdgeRepulsion / edgeCrossing;
        logger.info("Node/Edge-repulsion distance equated to crossing: "+Math.sqrt(nodeEdgeMinimumSq));

        prepareEdgesOfNodeMap();
        prepareEdgesCrossingMap();

        // Decrease temperature until endT reached
        while (T > endT) {
            if (T > edgeCrossCutoffT) {
                logger.info("T=" + T);
            } else  {
                logger.info("T=" + T + " (finetuning)");
            }
            // for each temperature, choose N times a node and
            // try to move it
            for (int i = 0; i < 30; i++) {
                for (Iterator it = g.nodes.iterator(); it.hasNext();) {
                    Node n = (Node) it.next();

                    tryAndMove(n);
                }
                Thread.yield();
            }
            T = gamma * T;
            if (c != null) {
                c.doLayout();
                //                c.repaint();
            }
            try {
                Thread.sleep(10);
            } catch (InterruptedException e) {
                logger.info("Aborted annealing");
                this.runner = null;
                return;
            }
        }
        this.runner = null;
        logger.info("Finished annealing layout");
    }

    public static void test() {
        Graph graph = new Graph();

        Node n1 = new AbstractNode("A", "A"); n1.setPosition(100, 100);
        Node n2 = new AbstractNode("B", "B"); n2.setPosition(200, 100);
        Node n3 = new AbstractNode("C", "C"); n3.setPosition(200, 200);
        Node n4 = new AbstractNode("D", "D"); n4.setPosition(100, 200);
        Node n5 = new AbstractNode("E", "E"); n5.setPosition(185, 175);

        graph.addEdge(new Edge(n1, n2));
        graph.addEdge(new Edge(n2, n3));
        graph.addEdge(new Edge(n3, n4));
        graph.addEdge(new Edge(n4, n1));
        graph.addEdge(new Edge(n1, n3));
        graph.addEdge(new Edge(n2, n4));
        graph.addEdge(new Edge(n1, n5));

        Frame f = new Frame("Testing GraphComponent");
        GraphComponent c = new GraphComponent(graph);

        f.add(c);
        f.setSize(600, 500);
        f.show();
        c.setShowLabels(NodeComponent.SHOW_LABEL_NAME);
        c.setZoomFixed(true);

        AnnealLayout layout = new AnnealLayout(graph, c);

        layout.setStartT(0.001);
        layout.setEndT(0.00000001);
        layout.setEdgeCrossCutoffT(0.1);
        layout.setEdgeStretch(0.0000001);
        layout.setEdgeCrossing(100);
        layout.setNodeEdgeRepulsion(1);
        layout.nodeEdgeMinimumSq = layout.nodeEdgeRepulsion / layout.edgeCrossing;
        logger.info("Minimum distance edge crossing: "+Math.sqrt(layout.nodeEdgeMinimumSq));

        // Calculate maxX, maxY
        layout.maxX = 0;
        layout.maxY = 0;
        for (Iterator i = layout.g.nodes.iterator(); i.hasNext();) {
            Node n = (Node) i.next();

            if (n.getX() > layout.maxX) layout.maxX = n.getX();
            if (n.getY() > layout.maxY) layout.maxY = n.getY();
        }
        layout.maxX *= 1.1;
        layout.maxY *= 1.1;

        layout.T = 0.001;
        layout.prepareEdgesOfNodeMap();

        double newX = 175;
        double newY = 175;

        double deltaE = layout.deltaE(n1, newX, newY);

        logger.info("deltaE   = " + deltaE);

        logger.info("Random test: " + Math.exp(-1.0 * deltaE / layout.T));

        n1.setPosition(175, 175);

        newX = 100;
        newY = 100;
        deltaE = layout.deltaE(n1, newX, newY);
        logger.info("deltaE back = " + deltaE);

        newX = 250;
        newY = 175;
        deltaE = layout.deltaE(n5, newX, newY);
        logger.info("deltaE for E out = " + deltaE);

        n5.setPosition(250,175);
        newX = 185;
        newY = 175;
        deltaE = layout.deltaE(n5, newX, newY);
        logger.info("deltaE for E back = " + deltaE);

        /*
         for(int i=0; i<100; i++) {
         layout.tryAndMove(n1);
         c.doLayout();
         //            c.repaint();
         }
         */
        c.doLayout();
        c.repaint();

        try {
            BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
            br.readLine();
        } catch (IOException e) {
        }

        layout.run();

        logger.info("Finished test");

        try {
            BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
            br.readLine();
        } catch (IOException e) {
        }
    }

}
