/**
 * Title:	Graph
 * Copyright:	Copyright (c) Jan-Pascal van Best
 * Email:	    janpascal@vanbest.org
 * Author:	    Jan-Pascal van Best
 * Web: 	http://www.tbm.tudelft.nl/webstaf/janb/software/graph/api/index.html
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.graph;


import java.util.*;
import java.io.*;
import java.lang.*;
import org.xml.sax.*;


/**
 * A mathematical graph, consisting of a set of nodes and a set of
 * edges between them. The nodes can be of any class type, but
 * implementing the {@link Node} <code>interface</code> is preferred.
 * The edges are of type {@link Pair}. You can add nodes and edges,
 * and when you add edges the related nodes are automatically added.
 * <p>
 * Graphs can be exported to
 * <a href="http://www.cwi.nl/InfoVisu/GraphXML/index.html">GraphXML</a>,
 * <a href="http://www.caida.org/tools/visualization/otter/">Otter</a>
 * and <a href="http://www.informatik.uni-bremen.de/daVinci/">daVinci</a>
 * (<a href="http://www.davinci-presenter.de/">new site</a>) file formats.
 * The {@link #removeDullNodes} method removes nodes with exactly one
 * outgoing and one incoming edge.
 *
 * @author	    Jan-Pascal van Best
 * @see 	Node
 * @see 	Pair
 */
public class Graph {

    /**
     * The nodes of this Graph. Elements can be of any type, but using a class
     * implementing {@link Node} has advantages when exporting the graph to
     * GraphXML.
     * Be sure to override the <code>hashCode</code>
     * and <code>equals</code> of the objects you use as nodes, to ensure that
     * duplicates are avoided.
     */
    public Set nodes;

    /**
     * The edges of this Graph, consisting of objects of type {@link Pair}.
     */
    public Set edges;

    /**
     * Constructs a new Graph with empty sets of nodes and edges.
     */
    public Graph() {
        nodes = new HashSet();
        edges = new HashSet();
    }

    /**
     * Adds a node to the graph. If the node is already in the graph,
     * it is not added again (since nodes is a set) and no warning is
     * given.
     *
     * @param	node	the node to be added
     * @see	Node
     */
    public void addNode(Object node) {
        nodes.add(node);
    }

    /**
     * Adds all nodes from newNodes to the graph. If any node is already in the graph,
     * it is not added again (since nodes is a set) and no warning is
     * given.
     *
     * @param	newNodes    the nodes to be added
     * @see	Node
     */
    public void addNodes(Collection newNodes) {
        nodes.addAll(newNodes);
    }

    /**
     * Adds an edge to the graph. If necessary, also adds the endpoints of
     * the edge to the graph as nodes.
     *
     * @param	edge	the edge to be added to the graph
     */
    public void addEdge(Edge edge) {
        if (!nodes.contains(edge.x)) {
            nodes.add(edge.x);
        }

        if (!nodes.contains(edge.y)) {
            nodes.add(edge.y);
        }
        edges.add(edge);
    }

    /**
     * Adds an edge from <code>from</code> to <code>to</code> to the graph.
     * If necessary, also adds <code>from</code> and <code>to</code> to
     * the graph as nodes.
     *
     * @param	from	the origin node of the edge to be added
     * @param	to	the destination node of the edge to be added
     */
    public void addEdge(Object from, Object to) {
        addEdge(new Edge(from, to));
    }

    /**
     * Calculates the outgoing edges of <code>node</code>. This means all edges in the
     * graph that originate in <code>node</code>, or all 
     * edges with <code>edge.x.equals(node)</code>. The edges are
     * returned in no particular order.
     *
     * @param	node	the node of which to calculate the outgoing edges
     * @return	a {@link Vector} consisting of the outgoing edges of
     *		<code>node</code>. If <code>node</code> is not in the graph,
     *		or if <code>node</code> has no outgoing edges, <code>outEdges</code>
     *		returns an empty <code>Vector</code>.
     */
    public Vector outEdges(Object node) {
        Vector result = new Vector();
        Iterator i = edges.iterator();

        while (i.hasNext()) {
            Edge edge = (Edge) i.next();

            if (edge.x.equals(node)) {
                result.add(edge);
            }
        } 
        return result;
    }

    /**
     * Calculates the incoming edges of <code>node</code>. This means all edges in the
     * graph that have <code>node</code> as destination, or all 
     * edges with <code>edge.y.equals(node)</code>. The edges are
     * returned in no particular order
     *
     * @param	node	the node of which to calculate the incoming edges
     * @return	a {@link Vector} consisting of the incoming edges of
     *		<code>node</code>. If <code>node</code> is not in the graph,
     *		or if <code>node</code> has no incoming edges, <code>inEdges</code>
     *		returns an empty <code>Vector</code>.
     */
    public Vector inEdges(Object node) {
        Vector result = new Vector();
        Iterator i = edges.iterator();

        while (i.hasNext()) {
            Edge edge = (Edge) i.next();

            if (edge.y.equals(node)) {
                result.add(edge);
            }
        }
        return result;
    }

    /**
     * return the subgraph with only the node sin subNodes and
     * all edges between those nodes (if there are also in this graph)
     */

    public Graph subGraph(Set subNodes) {
        Graph result = new Graph();
        result.addNodes(subNodes);
        for(Iterator i=edges.iterator(); i.hasNext(); ){
            Edge e = (Edge) i.next();
            if (subNodes.contains(e.x) && subNodes.contains(e.y)) {
                result.addEdge(e);
            }
        }
        return result;
    }

    /**
     * returns a new graph where all edge pairs (x,y) and (y,x) 
     * are replaced with a bidirectional edge (x,y)
     */
    public Graph mergeBidirEdges() {
        Graph result = new Graph();
        result.addNodes(nodes);
        Map newMap = new HashMap();
        for(Iterator i=edges.iterator(); i.hasNext(); ){
            Edge e = (Edge) i.next();
            Edge other = new Edge(e.y, e.x);
            if (edges.contains(other)) {
                if (result.edges.contains(other)) {
                    other = (Edge) newMap.get(other);
                    other.merge(e);
//                    System.out.println("Merged "+other+" to "+e);
                } else {
//                    System.out.println("Bidir edge between "+e.x+" and "+e.y);
                    Edge newEdge;
                    try {
                        newEdge = (Edge) e.clone();
                        newEdge.setBidir(true);
                        result.addEdge(newEdge);
                        newMap.put(e, newEdge);
//                        System.out.println("New edge: "+newEdge);
                    } catch (CloneNotSupportedException ex) {
                        ex.printStackTrace();
                    }
                }
            } else {
                result.addEdge(e);
            }
        }
        return result;
    }

	/**
	 * Get the maximum value over all edges of the property named <i>property</i>
	 */
	public double getMaxEdgeProperty(String property) {
		double max = -Double.MAX_VALUE;
        for(Iterator i=edges.iterator(); i.hasNext(); ){
            Edge e = (Edge) i.next();
	        String value = e.getProperty(property);
			try {
				double doubleValue = Double.parseDouble(value);
				if (doubleValue>max) max=doubleValue;
			} catch (NumberFormatException ex) {
				// Ignore value if not a number
			}
		}
		return max;
	}
	
/*
     public Iterator outEdgeIterator(Object node) {
     return new OutEdgeIterator(this, node);
     }

     public Iterator inEdgeIterator(Object node) {
     return new InEdgeIterator(this, node);
     }
     */

    public String getDTDExtension() {
        StringBuffer s = new StringBuffer();

        s.append("    <!ENTITY % nodeExtensions \"|property\">\n");
        s.append("    <!ELEMENT property EMPTY>\n");
        s.append("    <!ATTLIST property\n");
        s.append("        name    CDATA #REQUIRED\n");
        s.append("        content CDATA #IMPLIED\n");
        s.append("    >\n");
        return s.toString();
    }

    /**
     * Returns a <a href="http://www.cwi.nl/InfoVisu/GraphXML/index.html">GraphXML</a>
     * representation of a node. If <code>node</code> is an instance
     * of {@link Node}, the {@link Node#getName()}, {@link Node#getLabel()},
     * {@link Node#getPropertyNames()} and {@link Node#getProperty(java.lang.String)}
     * methods are used to fill in the XML representation. If <code>node</code> is
     * not an instance of {@link Node}, the {@link Object#toString} method is
     * used to get information about <code>node</code>.
     *
     * @param	node	the node which to get the XML representation
     * @return	a {@link Vector} consisting of the outgoing edges of
     *		<code>node</code>. If <code>node</code> is not in the graph,
     *		or if <code>node</code> has no outgoing edges, <code>outEdges</code>
     *		returns an empty <code>Vector</code>.
     */
    public String toXMLNodeString(Object node) {
        String s = "";

        if (node instanceof Node) {
            Node n = (Node) node;
            String name = n.getName().replace('&', '+');
            String label = n.getLabel().replace('&', '+');

            s = s + "        <node name=\"" + name + "\">\n";
            if (label != null && !label.equals("")) {
                s = s + "            <label>" + label + "</label>\n";
            }
            if ((n.getX() != Node.POSITION_UNKNOWN) && (n.getY() != Node.POSITION_UNKNOWN)) {
                s = s + "            <position x=\"" + n.getX() + "\" y=\"" + n.getY() + "\"/>\n";
            }
            List properties = n.getPropertyNames();

            for (Iterator i = properties.iterator(); i.hasNext();) {
                String pname = (String) i.next();
                String pvalue = n.getProperty(pname);

                if (pvalue != null) {
                    pvalue = pvalue.replace('&', '+');
                    s = s + "            <property name=\"" + pname + "\" content=\"" + pvalue + "\"/>\n";
                }
            }
            //	      s = s + " 	   <size width=\"10.0\" height=\"1.0\"/>\n";
            s = s + "        </node>\n";
        } else {
            String name = node.toString();

            name = name.replace('&', '+'); // XML does not like ampersands
            s = s + "        <node name=\"" + node + "\">\n";
            //	      s = s + " 	   <size width=\"10.0\" height=\"1.0\"/>\n";
            s = s + "        </node>\n";
        }
        return s;
    }

    public String toXMLEdgeString(Edge edge) {
        StringBuffer s = new StringBuffer(4000);

        String name = edge.getName();
        if (name != null) name = name.replace('&', '+');

        String label = edge.getLabel();
        if (label != null) label = label.replace('&', '+');

        String from = (edge.x instanceof Node) ? ((Node) edge.x).getName() : edge.x.toString();
        String to = (edge.y instanceof Node) ? ((Node) edge.y).getName() : edge.y.toString();
        if (from!=null) from = from.replace('&', '+'); // XML does not like ampersands
        if (to!=null) to = to.replace('&', '+');

        if (name != null && !name.equals("")) {
            s.append("        <edge name=\"" + name + "\" source=\"" + from + "\" target=\"" + to + "\">\n");
        } else {
            s.append("        <edge source=\"" + from + "\" target=\"" + to + "\">\n");
        }
        if (label != null && !label.equals("")) {
            s.append("            <label>" + label + "</label>\n");
        }

        /*
         p.println("	       <style>");
         p.println(" 	       <line colour=\"black\" linewidth=\"1.0\"/>");
         p.println(" 	   </style>");
         */
        List properties = edge.getPropertyNames();

        for (Iterator i = properties.iterator(); i.hasNext();) {
            String pname = (String) i.next();
            String pvalue = edge.getProperty(pname);

            if (pvalue != null) {
                pvalue = pvalue.replace('&', '+');
                s.append("            <property name=\"" + pname + "\" content=\"" + pvalue + "\"/>\n");
            }
        }
        s.append("        </edge>\n");

        return s.toString();
/*
        StringBuffer s = new StringBuffer();
        String from = (edge.x instanceof Node) ? ((Node) edge.x).getName() : edge.x.toString();
        String to = (edge.y instanceof Node) ? ((Node) edge.y).getName() : edge.y.toString();

        from = from.replace('&', '+'); // XML does not like ampersands
        to = to.replace('&', '+');
        s.append("        <edge source=\"" + from + "\" target=\"" + to + "\">\n");
        s.append("            <style>\n");
        s.append("                <line colour=\"black\" linewidth=\"1.0\"/>\n");
        s.append("            </style>\n");
        s.append("        </edge>\n");
        return s.toString();
*/
    }

    public String getExtraXMLElements() {
        return "";
    }
    
    /**
     * Writes a <a href="http://www.cwi.nl/InfoVisu/GraphXML/index.html">GraphXML</a>
     * representation of the graph to a disk file. The GraphXML definition is extended
     * (with extension mechanism defined in the GraphXML specification) to include
     * <code>&lt;property name="<i>name</i>" content="<i>content</i>"/&gt;</code>
     * tags in the nodes.
     * If (some of) the nodes of the graph are an instance
     * of {@link Node}, this method makes use of the contact of <code>Node</code>.
     *
     * @param	filename    the disk file name to which to write the XML.
     * @throws	IOException if something goes wrong with writing the file to disk.
     */
    public void exportXML(String filename) throws IOException {
        //	  System.out.println("Exporting GraphXML to "+filename);
        FileWriter f = new FileWriter(filename);
        PrintWriter p = new PrintWriter(f);

        p.println("<?xml version=\"1.0\"?>");
        String DTDExtension = getDTDExtension();

        if (DTDExtension != null && !DTDExtension.equals("")) {
            p.println("<!DOCTYPE GraphXML SYSTEM \"http://www.cwi.nl/InfoVisu/GraphXML/GraphXML.dtd\" [");
            p.print(DTDExtension);
            p.println("]>");
        } else {
            p.println("<!DOCTYPE GraphXML SYSTEM \"http://www.cwi.nl/InfoVisu/GraphXML/GraphXML.dtd\"");
        }
        p.println("<GraphXML>");
        p.println("    <graph isDirected=\"true\" isAcyclic=\"false\" id=\"test\">");
        for (Iterator i = nodes.iterator(); i.hasNext();) {
            String node = toXMLNodeString(i.next());

            p.println(node);
        }
        for (Iterator i = edges.iterator(); i.hasNext();) {
            String edge = toXMLEdgeString((Edge) i.next());

            p.println(edge);
        }
        p.println(getExtraXMLElements());
        p.println("    </graph>");
        p.println("</GraphXML>");
        p.flush();
        p.close();
        f.close();
    }

    public String toGraphMLNodeString(Object node) {
        String s = "";

        if (node instanceof Node) {
            Node n = (Node) node;
            String name = n.getName().replace('&', '+');
            String label = n.getLabel().replace('&', '+');

            s = s + "	     <node id=\"" + name + "\">\n";
            if (label != null && !label.equals("")) {
                s = s + "	     <data key=\"label\">" + label + "</data>\n";
            }
            if ((n.getX() != Node.POSITION_UNKNOWN) && (n.getY() != Node.POSITION_UNKNOWN)) {
                s = s + "	     <data key=\"pos\"><pos x=\"" + n.getX() + "\" y=\"" + n.getY() + "\"/></data>\n";
            }
            List properties = n.getPropertyNames();

            for (Iterator i = properties.iterator(); i.hasNext();) {
                String pname = (String) i.next();
                String pvalue = n.getProperty(pname);

                if (pvalue != null) {
                    pvalue = pvalue.replace('&', '+');
                    s = s + "		 <data key=\"" + pname + "\">" + pvalue + "</data>\n";
                }
            }
            s = s + "	     </node>\n";
        } else {
            String name = node.toString();

            name = name.replace('&', '+'); // XML does not like ampersands
            s = s + "	     <node id=\"" + name + "\">\n";
            s = s + "	     </node>\n";
        }
        return s;
    }

    public String toGraphMLEdgeString(Edge edge) {
        StringBuffer s = new StringBuffer();
        String from = (edge.x instanceof Node) ? ((Node) edge.x).getName() : edge.x.toString();
        String to = (edge.y instanceof Node) ? ((Node) edge.y).getName() : edge.y.toString();

        from = from.replace('&', '+'); // XML does not like ampersands
        to = to.replace('&', '+');
        s.append("	  <edge source=\"" + from + "\" target=\"" + to + "\"/>\n");
        return s.toString();
    }

    public void exportGraphML(String filename) throws IOException {
        //	  System.out.println("Exporting GraphML to "+filename);
        FileWriter f = new FileWriter(filename);
        PrintWriter p = new PrintWriter(f);

        p.println("<?xml version=\"1.0\"?>");
        p.println("<graphml>");
        p.println("    <graph edgedefault=\"directed\">");
        for (Iterator i = nodes.iterator(); i.hasNext();) {
            String node = toGraphMLNodeString(i.next());

            p.println(node);
        }
        for (Iterator i = edges.iterator(); i.hasNext();) {
            String edge = toGraphMLEdgeString((Edge) i.next());

            p.println(edge);
        }
        //	  p.println(getExtraXMLElements());
        p.println("    </graph>");
        p.println("</graphml>");
        p.flush();
        p.close();
        f.close();
    }

    /**
     * Writes a <a href="http://www.informatik.uni-bremen.de/daVinci/">daVinci</a>
     * representation of the graph to a disk file.
     *
     * @param	filename    the disk file name to which to write the daVinci file
     * @throws	IOException if something goes wrong with writing the file to disk.
     */
    public void exportdaVinci(String filename) throws IOException {
        //	  System.out.println("Exporting daVinci data to "+filename);
        PrintWriter p = new PrintWriter(new FileWriter(filename));

        p.println("[");
        for (Iterator i = nodes.iterator(); i.hasNext();) {
            Object node = i.next();

            p.println("    l(\"" + node.toString() + "\",n(\"\",[a(\"OBJECT\",\"" + node.toString() + "\")],[");
            Iterator children = outEdges(node).iterator();

            while (children.hasNext()) {
                Edge edge = (Edge) children.next();

                p.println("	   l(\"" + edge.x.toString() + "->" + edge.y.toString() + "\",e(\"\",[],");
                if (children.hasNext())
                    p.println(" 	   r(\"" + edge.y.toString() + "\"))),");
                else
                    p.println(" 	   r(\"" + edge.y.toString() + "\")))");
            }
            if (i.hasNext()) p.println("    ])),");
            else p.println("	]))");
        }
        p.println("]");
        p.close();
    }

    /**
     * Writes an <a href="http://www.caida.org/tools/visualization/otter/">Otter</a>
     * representation of the graph to a disk file. 
     *
     * @param	filename    the disk file name to which to write the Otter file
     * @throws	IOException if something goes wrong with writing the file to disk.
     */
    public void exportOtter(String filename) throws IOException {
        //	  System.out.println("Exporting to Otter file "+filename);
        PrintWriter p = new PrintWriter(new FileWriter(filename));

        p.println("t " + nodes.size()); // total number of nodes
        p.println("T " + edges.size()); // total number of links

        // Two groups of properties: Host, TestBox, AutSystem
        p.println("g 0 s 4 Host Properties");
        p.println("f 0 IP address'host name'AS number'AS name");

        p.println("g 1 s 2 Testbox Properties");
        p.println("f 1 Testbox Number'Description");

        p.println("g 2 s 2 Autonomous System Properties");
        p.println("f 2 AS number'AS name");

        Hashtable nodesTable = new Hashtable(); // this table allows us to find the index number of a node

        int index = 0;

        for (Iterator i = nodes.iterator(); i.hasNext(); index++) {
            Object n = i.next();

            nodesTable.put(n, new Integer(index));
            if (n instanceof Node) {
                Node node = (Node) n;
                String latitude = node.getProperty("latitude");
                String longitude = node.getProperty("longitude");

                if (latitude != null && longitude != null)
                    p.println("N " + index + " " + latitude + " " + longitude + " " + node.getLabel()); 
                else
                    p.println("? " + index + " " + node.getLabel());

                String hostName = node.getProperty("hostname");

                if (hostName != null) {
                    // we've got a host
                    p.println("v " + index + " 0 " + node.getProperty("IPaddress") + "'" + hostName + "'" + node.getProperty("ASnumber") + "'" + node.getProperty("ASname"));
                }

                String testboxnumber = node.getProperty("testboxnumber");

                if (testboxnumber != null) {
                    // we've got a test box
                    p.println("v " + index + " 1 " + testboxnumber + "'" + node.getProperty("description"));
                }

                String ASnumber = node.getProperty("ASnumber");

                if (ASnumber != null) {
                    // we've got a test box
                    p.println("v " + index + " 2 " + ASnumber + "'" + node.getProperty("ASname"));
                }

            } else
                p.println("? " + index + " " + n.toString());
        }

        index = 0;
        for (Iterator i = edges.iterator(); i.hasNext(); index++) {
            Edge edge = (Edge) i.next();

            // unidirectional link: l linkidx n_idx1 n_idx2
            p.println("l " + index + " " + nodesTable.get(edge.x) + " " + nodesTable.get(edge.y));
        }

        p.close();
    }

    /**
     * Removes all dull nodes from the graph. A dull node is a node with
     * exactly one incoming edge and exactly one outgoing edge. In the following
     * drawing, <code>B</code> is a dull node:
     * <code>A ---> B ---> C</code>
     * When a dull node is removed, its incoming and outgoing edges are replaced
     * by a new edge from the node originating node of the incoming edge to
     * the destination node of the outgoing edge, in this case an edge from
     * <code>A</code> to <code>C</code>.
     */
    // 
    public void removeDullNodes() {
        boolean finished = false;

        while (!finished) {
            finished = true;
            for (Iterator i = nodes.iterator(); i.hasNext();) {
                Object node = i.next();
                Vector in = inEdges(node);

                if (in.size() == 1) {
                    Vector out = outEdges(node);

                    if (out.size() == 1) {
                        // we've found a dull node, remove it and relink
                        Edge from = (Edge) in.get(0);
                        Edge to = (Edge) out.get(0);

                        edges.add(new Edge(from.x, to.y));
                        edges.remove(from);
                        edges.remove(to);
                        i.remove(); // remove current node from iterator and from nodes
                        finished = false;
                    }
                }
            }
        }
    }

    /**
     * Returns a string representation of this graph. The string representation
     * consists of a listing of the nodes and edges of the graph.
     *
     * @return a string representing the graph
     */
    public String toString() {
        StringBuffer s = new StringBuffer();

        s.append("Nodes: ");
        for (Iterator i = nodes.iterator(); i.hasNext();) {
            s.append(i.next().toString() + ",");
        }
        s.append("\nEdges: ");
        for (Iterator i = edges.iterator(); i.hasNext();) {
            s.append(i.next().toString() + ",");
        }
        return s.toString();
    }

}

