/**
 * Title:		Graph
 * Copyright:	Copyright (c) Jan-Pascal van Best
 * Email:	    janpascal@vanbest.org
 * Author:	    Jan-Pascal van Best
 * Web: 		http://www.tbm.tudelft.nl/webstaf/janb/software/graph/api/index.html
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.graph;

import java.io.*;
import java.lang.*;
import java.util.*;
import java.util.logging.*;

import java.awt.Color;

public class PartialLayout implements Layout {
    public final static int GET_ALL_NODES = 0;
    public final static int GET_LOW_DEGREE_NODES = 1;
    public final static int GET_HIGH_DEGREE_NODES = 2;

    public final static String[] PARTIAL_STRINGS = 
        { "All nodes", "Low-degree nodes", "High-degree nodes" };

    private Graph g;
    private GraphComponent c;
    private Layout layout;
    double part = 0.8;
    int whichPart = GET_LOW_DEGREE_NODES;
    boolean mergeBidir = false;
    private static Logger logger = Logger.getLogger("nl.tudelft.tbm.graph");

    public PartialLayout(Graph g, GraphComponent c, Layout layout, double part, int whichPart, boolean mergeBidir) {
        this.g = g;
        this.c = c;
        this.layout = layout;
        this.part = part;
        this.whichPart = whichPart;
        this.mergeBidir = mergeBidir;
    }

    public PartialLayout(Graph g, Layout layout) {
        this.g = g;
        this.c = null;
    }

    public void setGraph(Graph g, GraphComponent c) {
        this.g = g;
        this.c = c;
    }

    private Graph calcSubGraph(Graph g) {
        if (mergeBidir) {
            logger.info("Merging bidirectional edges");
            g = g.mergeBidirEdges();
        }

        if(whichPart==GET_ALL_NODES) {
            return g;
        }

        int numNodes = g.nodes.size();
        int[] degree = new int[numNodes];
        Node[] nodes = new Node[numNodes];

        // Nodemap maps nodes to node numbers
        Map nodeMap = new HashMap();
        int count = 0;
        for (Iterator i = g.nodes.iterator(); i.hasNext();) {
            Node n = (Node) i.next();
            nodeMap.put(n, new Integer(count));
            nodes[count] = n;
            count++;
        }
        logger.info("numNodes: "+numNodes);

        // degree[i] is the degree of node #i
        for (Iterator i = g.edges.iterator(); i.hasNext();) {
            Edge e = (Edge) i.next();
            int nodeNum = ((Integer) nodeMap.get(e.x)).intValue();
            degree[nodeNum]++;
            nodeNum = ((Integer) nodeMap.get(e.y)).intValue();
            degree[nodeNum]++;
        }

        // max is the maximum degree of all nodes
        int max = 0;
        for(int i=0; i<numNodes; i++) {
            if (degree[i]>max) max=degree[i];
        }
        max++;

        logger.info("max degree: "+max);
        // histogram[i] is the number of nodes with degree i
        int[] histogram = new int[max];
        for(int i=0; i<numNodes; i++) {
            histogram[degree[i]]++;
        }
        StringBuffer s = new StringBuffer();
        s.append("Degree histogram: ");
        for(int i=0; i<histogram.length; i++) {
            s.append(histogram[i]+",");
        }
        logger.info(s.toString());

        Set subNodes = new HashSet();
        long numIncluded = Math.round(numNodes * part);
        int runningTotal = 0;
        int threshold = 0;
        switch(whichPart) {
        case GET_LOW_DEGREE_NODES:
            logger.info("Getting low degree nodes");
            for(int i=0; i<max; i++) {
                runningTotal += histogram[i];
                if (runningTotal>=numIncluded) {
                    threshold = i+1;
                    break;
                }
            }

            logger.info("threshold: "+threshold);
            // Select nodes to include in result
            for (int i = 0; i<numNodes; i++) {
                if (degree[i]<threshold) {
                    subNodes.add(nodes[i]);
                } else {
                    // Also color unselected nodes yellow
                    if (c != null) {
                        NodeComponent nc = (NodeComponent) c.nodeMap.get(nodes[i]);
                        nc.setColor(Color.YELLOW);
                        nc.repaint();
                    }
                }
            }
            break;
        case GET_HIGH_DEGREE_NODES:
            logger.info("Getting high degree nodes");
            for(int i=max-1; i>=0; i--) {
                runningTotal += histogram[i];
                if (runningTotal>=numIncluded) {
                    threshold = i;
                    break;
                }
            }

            logger.info("threshold: "+threshold);
            // Select nodes to include in result
            for (int i = 0; i<numNodes; i++) {
                if (degree[i]>=threshold) {
                    subNodes.add(nodes[i]);
                } else {
                    // Also color unselected nodes yellow
                    if (c != null) {
                        NodeComponent nc = (NodeComponent) c.nodeMap.get(nodes[i]);
                        nc.setColor(Color.YELLOW);
                        nc.repaint();
                    }
                }
            }
        }

        logger.info("Selected "+subNodes.size()+" nodes");
        
        // and hide unselected edges
        if (c != null) {
            for (Iterator i = g.edges.iterator(); i.hasNext();) {
                Edge e = (Edge) i.next();
                if ( ! (subNodes.contains(e.x) && subNodes.contains(e.y)) ) {
                    EdgeComponent ec = (EdgeComponent) c.edgeMap.get(e);
                    ec.hide();
                    // also hide reverse edge, in available
                    Edge other = new Edge(e.y, e.x);
                    ec = (EdgeComponent) c.edgeMap.get(other);
                    if (ec != null) ec.hide();
                }
            }
        }

        // Move unselected nodes to the right
        int numUnselected = numNodes - subNodes.size();
        if (c != null && numUnselected > 0) {
            double maxX = (c.getWidth() - (2 * c.NODE_SIZE) - 50) / (c.factorX);
            double maxY = (c.getHeight() - (2 * c.NODE_SIZE) - 50) / (c.factorY);
            double unselDistance = maxY / (numUnselected+1);
            int unselNumber = 1;
            for (int i = 0; i<numNodes; i++) {
                if (!subNodes.contains(nodes[i])) {
                    nodes[i].setX(maxX*1.05);
                    nodes[i].setY(unselNumber * unselDistance);
                    unselNumber++;
                }
            }
        }

        // Take subgraph
        Graph result = g.subGraph(subNodes);
        
        return result;
    }

    public void doLayout() {
        Graph subGraph = calcSubGraph(g);
        c.doLayout();

        layout.setGraph(subGraph, c);
        layout.doLayout();
    }

    public void interruptLayout() {
        if (layout != null) {
            layout.interruptLayout();
        }
    }

    public boolean isLayouting() {
        if (layout != null) {
            return layout.isLayouting();
        } else {
            return false;
        }
    }
}
